use std::{
    env,
    error::Error,
    io::{ErrorKind, Read, Write},
    net::{Shutdown, TcpListener, TcpStream},
    sync::{Arc, Mutex},
    thread,
};

use rustls::{
    ClientConfig, ClientConnection, DigitallySignedStruct, RootCertStore, SignatureScheme,
    StreamOwned,
    client::danger::{HandshakeSignatureValid, ServerCertVerified, ServerCertVerifier},
    pki_types::{CertificateDer, ServerName, UnixTime},
};
use threadpool::ThreadPool;

#[derive(Debug)]
pub struct NoCertVerify;

impl ServerCertVerifier for NoCertVerify {
    fn verify_server_cert(
        &self,
        _: &CertificateDer<'_>,
        _: &[CertificateDer<'_>],
        _: &ServerName<'_>,
        _: &[u8],
        _: UnixTime,
    ) -> Result<ServerCertVerified, rustls::Error> {
        Ok(ServerCertVerified::assertion())
    }

    fn verify_tls12_signature(
        &self,
        _: &[u8],
        _: &CertificateDer<'_>,
        _: &DigitallySignedStruct,
    ) -> Result<HandshakeSignatureValid, rustls::Error> {
        Ok(HandshakeSignatureValid::assertion())
    }

    fn verify_tls13_signature(
        &self,
        _: &[u8],
        _: &CertificateDer<'_>,
        _: &DigitallySignedStruct,
    ) -> Result<HandshakeSignatureValid, rustls::Error> {
        Ok(HandshakeSignatureValid::assertion())
    }

    fn supported_verify_schemes(&self) -> Vec<SignatureScheme> {
        vec![
            SignatureScheme::ECDSA_NISTP256_SHA256,
            SignatureScheme::ECDSA_NISTP384_SHA384,
            SignatureScheme::ECDSA_NISTP521_SHA512,
            SignatureScheme::ECDSA_SHA1_Legacy,
            SignatureScheme::ED25519,
            SignatureScheme::ED448,
            SignatureScheme::RSA_PKCS1_SHA1,
            SignatureScheme::RSA_PKCS1_SHA256,
            SignatureScheme::RSA_PKCS1_SHA384,
            SignatureScheme::RSA_PKCS1_SHA512,
            SignatureScheme::RSA_PSS_SHA256,
            SignatureScheme::RSA_PSS_SHA384,
            SignatureScheme::RSA_PSS_SHA512,
        ]
    }
}

fn accept_client(
    stream: Arc<TcpStream>,
    config: Arc<ClientConfig>,
    sni_domain: String,
    remote_stream: Arc<TcpStream>,
) -> Result<(), Box<dyn Error>> {
    stream.set_nonblocking(true)?;
    remote_stream.set_nonblocking(true)?;

    let mut tls = StreamOwned::new(
        ClientConnection::new(config, sni_domain.try_into()?)?,
        remote_stream.try_clone()?,
    );

    while tls.conn.is_handshaking() {
        match tls.conn.complete_io(&mut tls.sock) {
            Ok(_) => {},
            Err(e) => {
                if e.kind() != ErrorKind::WouldBlock {
                    return Ok(());
                }
            }
        }
    }

    let tls = Arc::new(Mutex::new(tls));

    thread::spawn({
        let tls = tls.clone();
        let stream = stream.clone();

        move || {
            loop {
                let mut buffer = vec![0; 4096];
                match (&*stream).read(&mut buffer) {
                    Ok(0) => { break },
                    Ok(n) => {
                        // println!("from remote {n}");
                        buffer.truncate(n);
                        match tls.lock().unwrap().write(&buffer) {
                            Ok(_) => {},
                            Err(e) => {
                                if e.kind() != ErrorKind::WouldBlock {
                                    break;
                                }
                            }
                        }
                    }
                    Err(e) => {
                        if e.kind() != ErrorKind::WouldBlock {
                            break;
                        }
                    }
                }
            }
        }
    });

    let tls = tls.clone();
    let stream = stream.clone();

    loop {
        let mut buffer = vec![0; 4096];
        match tls.lock().unwrap().read(&mut buffer) {
            Ok(0) => { break },
            Ok(n) => {
                // println!("from tls {n}");
                buffer.truncate(n);
                match (&*stream).write(&buffer) {
                    Ok(_) => {},
                    Err(e) => {
                        if e.kind() != ErrorKind::WouldBlock {
                            return Err(e.into());
                        }
                    }
                }
            }
            Err(e) => {
                if e.kind() != ErrorKind::WouldBlock {
                    return Err(e.into());
                }
            }
        }
    }

    Ok(())
}

fn main() -> Result<(), Box<dyn Error>> {
    let mut args = env::args();

    args.next();

    let sni_domain = args.next().expect("missing sni domain argument");
    let host = args.next().expect("missing host argument");
    let local_host = args.next().expect("missing local host argument");

    let mut config = ClientConfig::builder()
        .with_root_certificates(RootCertStore::empty())
        .with_no_client_auth();

    let verifier = Arc::new(NoCertVerify);
    config.dangerous().set_certificate_verifier(verifier);

    let config = Arc::new(config);

    let threadpool = ThreadPool::new(10);

    let listener = TcpListener::bind(local_host).unwrap();

    for stream in listener.incoming() {
        let stream = stream.expect("listener got broken");

        threadpool.execute({
            let config = config.clone();
            let sni_domain = sni_domain.clone();
            let host = host.clone();

            move || {
                let Ok(remote_stream) = TcpStream::connect(host) else {
                    return;
                };

                let stream = Arc::new(stream);
                let remote_stream = Arc::new(remote_stream);

                if let Err(e) = accept_client(stream.clone(), config, sni_domain, remote_stream) {
                    println!("error connection: {e:?}");
                }

                let _ = (&*stream).shutdown(Shutdown::Both);
            }
        });
    }

    Ok(())
}
